from control.controller_decentralized import Decentralized
from tqdm import tqdm

class DecentralizedRL(Decentralized):

    def __init__(self, env, agents):
        # initialize super class
        super().__init__(env, agents)

    def run(self, render=False, max_iteration=None, max_episode_len=1, num_episodes=1, batch_size=0):
        """Runs the controller on the environment given in the init,
        with the agents given in the init

        Args:
            render (bool, optional): Whether to render while runngin. Defaults to False.
            max_iteration ([type], optional): Number of steps to run. Defaults to infinity.
            max_episode_len (int, optional): Upper cap on episode length. Defaults to 1.
            num_episodes (int, optional): Number of episodes. Defaults to 1.
            batch_size (int, optional): Batch size for the training algorithm. Defaults to 0.
        """

        # print("Training...")
        # self.train(max_episode_len, num_episodes, batch_size=batch_size)
        # print("Finished Training")
        done = False
        index = 0
        observation = self.environment.get_env().reset()
        self.total_rewards = []
        while done is not True:
            index += 1
            if max_iteration is not None and index > max_iteration:
                break

            # display environment
            if render:
                self.environment.get_env().render()

            # get actions for each agent to perform

            joint_action = self.get_joint_action(observation)
            observation, reward, done, info = self.perform_joint_action(joint_action)
            self.total_rewards.append(reward)
            done = all(value == True for value in done.values())
            if done:
                break

        if render:
            self.environment.get_env().render()

    def train(self, max_episode_len, num_episodes, batch_size=0):
        """Train the agents

        Args:
            max_episode_len (int, optional): Upper cap on episode length. Defaults to 1.
            num_episodes (int, optional): Number of episodes. Defaults to 1.
            batch_size (int, optional): Batch size for the training algorithm. Defaults to 0.
        """
        episode_rewards = [0.0]

        print("Starting training...")
        for i in tqdm(range(num_episodes)):
            obs = self.environment.get_env().reset()
            ep_reward = 0.0
            agent_rewards = {agent_name: 0.0 for agent_name in self.agents.keys()}
            for _ in range(max_episode_len):
                actions = {agent_name: self.agents[agent_name].get_decision_maker().get_train_action(obs[agent_name]) for
                               agent_name in self.agents.keys()}
                new_obs, rewards, done, info = self.environment.get_env().step(actions)
                for agent_name in self.agents.keys():
                    self.agents[agent_name].get_decision_maker().update_step(obs[agent_name], actions[agent_name],
                                                            new_obs[agent_name], rewards[agent_name], done[agent_name])
                    ep_reward += rewards[agent_name]

                if batch_size > 0:
                    for agent_name in self.agents.keys():
                        self.agents[agent_name].get_decision_maker().update_episode(batch_size)

                obs = new_obs
                terminal = False
                terminal = all(value == True for value in done.values())

                if terminal:
                    break

            episode_rewards.append(ep_reward)

            for agent_name in self.agents.keys():
                self.agents[agent_name].get_decision_maker().update_episode()



from agents.agent import Agent

class RandomDecisionMaker:
    def __init__(self, action_space):
        self.space = action_space

    def get_action(self, observation):
        if type(self.space) == dict:
            return {agent: self.space[agent].sample() for agent in self.space.keys()}
        else:
            return self.space.sample()

class RandomAgent(Agent):
    """action_space can be either an action space (Discrete, Box etc)
       or a dictionary (key per agent) of action spaces
    """
    def __init__(self, action_space):
        self.decision_maker = RandomDecisionMaker(action_space)
    



import time

import numpy as np

class TaxiTaskAgent():
    """Like Goal agent, also also bids and the goal
    can be set later
    """
    def __init__(self, position=[0, 0]):
        self.pos = position # We need to know to position for the bidding:

    def reset(self, observation):
        self.pos = tuple(observation[0][:2])
    
    def get_bid(self, suggested_task):
        # Manhatten Distances Sum to the passenger and then to the destination
        manhatten = lambda pos1, pos2: abs(pos1[0] - pos2[0]) + abs(pos1[1] - pos2[1])
        pass_loc, dest_loc = suggested_task
        return manhatten(self.pos, pass_loc) + manhatten(pass_loc, dest_loc)

    def set_task(self, task):
        self.pass_loc, self.dest_loc = task

    def get_action(self, observation):
        return 0

    def navigate(self, dest):
        pass


from abc import ABC, abstractmethod

class Controller(ABC):
    """An abstract controller class, for other controllers
    to inherit from
    """

    # init agents and their observations
    def __init__(self, environment, agents, central_agent=None):
        self.environment = environment
        self.agents = agents
        self.central_agent = central_agent

    def run(self, render=False, max_iteration=None):
        """Runs the controller on the environment given in the init,
        with the agents given in the init

        Args:
            render (bool, optional): Whether to render while runngin. Defaults to False.
            max_iteration ([type], optional): Number of steps to run. Defaults to infinity.
        """
        done = False
        index = 0
        observation = self.environment.get_env().reset()
        self.total_rewards = []
        while done is not True:
            index += 1
            if max_iteration is not None and index > max_iteration:
                break

            # display environment
            if render:
                self.environment.get_env().render()

            # get actions for each agent to perform

            joint_action = self.get_joint_action(observation)
            observation, reward, done, info = self.perform_joint_action(joint_action)
            self.total_rewards.append(reward)
            done = all(value == True for value in done.values())
            if done:
                break

        if render:
            self.environment.get_env().render()

    def perform_joint_action(self, joint_action):
        return self.environment.get_env().step(joint_action)

    def get_joint_action(self, observation):
        pass



from control.controller import Controller


class Decentralized(Controller):

    def __init__(self, env, agents):
        # initialize super class
        super().__init__(env, agents)

    def get_joint_action(self, observation):
        """Returns the joint action

        Args:
            observation (dict): the current observatins

        Returns:
            dict: the actions for the agents
        """
        joint_action = {}
        for agent_name in self.agents.keys():
            action = self.agents[agent_name].get_decision_maker().get_action(observation[agent_name])
            joint_action[agent_name] = action

        return joint_action


from control.controller_centralized import Centralized
import numpy as np
from tqdm import tqdm

"""Class to use when using centralized controller
for RL setting
"""
class CentralizedRL(Centralized):

    def __init__(self, env, agents, central_agent):
        # initialize super class
        super().__init__(env, agents, central_agent)

    def run(self, render=False, max_iteration=None, max_episode_len=1, num_episodes=1, batch_size=0):
        """Runs the controller on the environment given in the init,
        with the agents given in the init

        Args:
            render (bool, optional): Whether to render while runngin. Defaults to False.
            max_iteration ([type], optional): Number of steps to run. Defaults to infinity.
            max_episode_len (int, optional): Upper cap on episode length. Defaults to 1.
            num_episodes (int, optional): Number of episodes. Defaults to 1.
            batch_size (int, optional): Batch size for the training algorithm. Defaults to 0.
        """
        # print("Training...")
        # self.train(max_episode_len, num_episodes, batch_size=batch_size)
        # print("Finished Training")

        done = False
        index = 0
        observation = self.environment.get_env().reset()
        self.total_rewards = []
        while done is not True:
            index += 1
            if max_iteration is not None and index > max_iteration:
                break

            # display environment
            if render:
                self.environment.get_env().render()

            # get actions for each agent to perform

            joint_action = self.get_joint_action(observation)
            observation, reward, done, info = self.perform_joint_action(joint_action)
            self.total_rewards.append(reward)
            done = all(value == True for value in done.values())
            if done:
                break

        if render:
            self.environment.get_env().render()

    def get_joint_action(self, observation):
        """Returns the joint actions of all the agents

        Args:
            observation ([dict]): The agents observations

        Returns:
            dict: dict of all the actions
        """
        observations = []
        # Dict to list:
        for agent_name in self.agents.keys():
            observations.append(observation[agent_name])

        state = self.decode_state(observations, self.environment.get_needs_conv())
        joint_act = self.decision_maker.get_action(state)
        joint_act = self.decode_action(joint_act, self.environment.get_num_actions(),
                                       len(self.environment.get_env_agents()))
        joint_action = {}
        for i, agent_name in enumerate(self.environment.get_env_agents()):
            action = joint_act[i]
            joint_action[agent_name] = action

        return joint_action

    def decode_state(self, obs, needs_conv):
        """Turns the ovservation from a list to np array

        Args:
            obs (list): list of observations
            needs_conv (bool): whether we want conv layers (affects the shape)

        Returns:
            ndarray: the observations
        """
        if needs_conv:
            return np.vstack(obs)
        else:
            return np.hstack(obs)

    def decode_action(self, action, num_actions, num_agents):
        """Decodes the action from the model to RL env friendly format

        Args:
            action (int): The action from the model
            num_actions (int): number of actions avaiable to every agent
            num_agents (int): number of agents

        Returns:
            list: list of individual actions
        """
        out = []
        for ind in range(num_agents):
            out.append(action % num_actions)
            action = action // num_actions
        return list(reversed(out))

    def train(self, max_episode_len, num_episodes, batch_size=0):
        """Train the agents

        Args:
            max_episode_len (int, optional): Upper cap on episode length. Defaults to 1.
            num_episodes (int, optional): Number of episodes. Defaults to 1.
            batch_size (int, optional): Batch size for the training algorithm. Defaults to 0.
        """
        episode_rewards = [0.0]

        print("Starting training...")
        for i in tqdm(range(num_episodes)):
            obs = self.environment.get_env().reset()
            ep_reward = 0.0
            for _ in range(max_episode_len):

                observations = []
                for agent_name in self.agents.keys():
                    observations.append(obs[agent_name])

                state = self.decode_state(observations, self.environment.get_needs_conv())
                joint_act = self.decision_maker.get_train_action(state)
                decoded_joint_act = self.decode_action(joint_act, self.environment.get_num_actions(),
                                               len(self.environment.get_env_agents()))

                actions = {}
                for i, agent_name in enumerate(self.environment.get_env_agents()):
                    action = decoded_joint_act[i]
                    actions[agent_name] = action

                new_obs, rewards, done, info = self.environment.get_env().step(actions)

                new_observations = []
                reward = 0
                for agent_name in self.agents.keys():
                    new_observations.append(new_obs[agent_name])
                    reward += rewards[agent_name]

                new_state = self.decode_state(observations, self.environment.get_needs_conv())
                new_done = all(value == True for value in done.values())

                self.decision_maker.update_step(state, joint_act, new_state, reward, new_done)
                ep_reward += reward

                if batch_size > 0:
                    self.decision_maker.update_episode(batch_size)

                obs = new_obs
                terminal = False
                terminal = new_done

                if terminal:
                    break

            episode_rewards.append(ep_reward)

            self.decision_maker.update_episode()


import sys
import itertools
import numpy as np
import matplotlib.pyplot as plt
sys.path.append('../src')
from control.controller_decentralized import Decentralized
from control.controller_decentralized_RL import DecentralizedRL
from control.controller_centralized_RL import CentralizedRL
from agents.deep_policy_gradient import DeepPolicyGradient
from agents.policy_gradient import PolicyGradient
from agents.deep_q import DQN
from agents.agent import Agent
from agents.random_agent import RandomAgent
from environments.env_wrapper import EnvWrappper

MAX_EPISODE_LEN = 25

def main():
    mode = 'CENT_DPG'
    
    if mode == 'DECENT_RANDOM':
        controller = make_random_decentralized_control()
        rl = False

    elif mode == 'DECENT_DPG':
        controller = make_rl_decentralized_control('dpg')
        controller.train(MAX_EPISODE_LEN, 1000)
        rl = True

    elif mode == 'CENT_DPG':
        controller = make_rl_centralized_control('dpg')
        controller.train(MAX_EPISODE_LEN, 1000)
        rl = True
    
    results = evaluate_controller(controller, 100, RL=rl)
    sums = sum_all_agents(results)
    plt.style.use('seaborn')
    plt.hist(sums)
    print(f'Mean is {sum(sums)/len(sums)}')
    plt.show()

def make_rl_centralized_control(centralized_policy):
    env = set_env()
    decision_maker = create_centralized_agent(centralized_policy, env)
    env_agents = env.get_env_agents()
    centralized_agents = {agent_name: None  for agent_name in env_agents}
    controller = CentralizedRL(env, centralized_agents, decision_maker)
    return controller

def make_rl_decentralized_control(decentralized_policy):
    env = set_env()
    env_agents = env.get_env_agents()
    decentralized_agents = {agent_name: create_decentralized_agent(decentralized_policy, env)
          for agent_name in env_agents}
    controller = DecentralizedRL(env, decentralized_agents)
    return controller

def make_random_decentralized_control():
    env = set_env()
    # Make Random Decentralized
    spaces = env.get_env().action_spaces
    
    agents = {
        agent_name: RandomAgent(spaces[agent_name])
        for agent_name in spaces
    }
    controller = Decentralized(env, agents)
    return controller

def evaluate_controller(controller, num_runs, RL):
    all_results = []
    for _ in range(num_runs):
        if RL:
            controller.run(render=False, max_iteration=10, max_episode_len=25, num_episodes=10, batch_size=0)
        else:
            controller.run(False, 20)
        all_results.append(controller.total_rewards)
    return all_results

def set_env():
    sys.path.append('../environments/MultiTaxiEnv')
    from environments.MultiTaxiEnv.taxi_environment import TaxiEnv
    env = TaxiEnv(2)
    # Make sure it works with our API:
    env.agents = env.taxis_names
    env.action_spaces = {
        agent_name: env.action_space for agent_name in env.agents
    }
    env.observation_spaces = {
        agent_name: env.observation_space for agent_name in env.agents
    }
    env.possible_agents = [agent for agent in env.agents]
    needs_conv = False
        
    return EnvWrappper(env, needs_conv=needs_conv)
    
def sum_all_agents(results):
    sums = [] # Across all agents
    for episode_results in results:
        agents_sums = list(itertools.chain.from_iterable([
            step_results.values() for step_results in episode_results
        ]))
        sums.append(sum(agents_sums))
    return sums

def create_centralized_agent(policy_name, env):
    needs_conv = env.get_needs_conv()
    num_obs = env.get_num_obs() if needs_conv else\
        (1, env.get_num_obs()[::-1][0] * (len(env.get_env_agents())))
    num_actions = (env.get_num_actions()) ** (len(env.get_env_agents()))    
    mapping_fn = lambda x: x.flatten() if not needs_conv else None
    
    if policy_name == 'pg':
        return Agent(PolicyGradient(num_actions, num_obs, mapping_fn=mapping_fn))
    
    elif policy_name == 'dpg':
        return Agent(DeepPolicyGradient(num_obs, num_actions, is_conv=needs_conv,
                                        mapping_fn=mapping_fn)) 
        
    elif policy_name == 'dqn':
        return Agent(DQN(num_obs, num_actions, is_conv=needs_conv,
                                        mapping_fn=mapping_fn)) 
        
    print("Invalid Policy!")       
    return

def create_decentralized_agent(policy_name, env):
    num_obs = env.get_num_obs()
    num_actions = env.get_num_actions()
    needs_conv = env.get_needs_conv()
    mapping_fn = lambda x: x.flatten() if not needs_conv else None
    
    if policy_name == 'pg':
        return Agent(PolicyGradient(num_actions, num_obs, mapping_fn=mapping_fn))
    
    elif policy_name == 'dpg':
        return Agent(DeepPolicyGradient(num_obs, num_actions, is_conv=needs_conv,
                                        mapping_fn=mapping_fn)) 
        
    elif policy_name == 'dqn':
        return Agent(DQN(num_obs, num_actions, is_conv=needs_conv,
                                        mapping_fn=mapping_fn)) 
        
    print("Invalid Policy!")       
    return

if __name__ == "__main__":
    main()


import cv2
import os
import matplotlib.pyplot as plt
import numpy as np


def save_img(rgb_arr, path, name):
    plt.imshow(rgb_arr, interpolation='nearest')
    plt.savefig(path + name)


def make_video_from_image_dir(vid_path, img_folder, video_name='trajectory', fps=5):
    """
    Create a video from a directory of images
    """
    images = [img for img in os.listdir(img_folder) if img.endswith(".png")]
    images.sort()

    rgb_imgs = []
    for i, image in enumerate(images):
        img = cv2.imread(os.path.join(img_folder, image))
        rgb_imgs.append(img)

    make_video_from_rgb_imgs(rgb_imgs, vid_path, video_name=video_name, fps=fps)


def make_video_from_rgb_imgs(rgb_arrs, vid_path, video_name='trajectory',
                             fps=5, format="mp4v", resize=(640, 480)):
    """
    Create a video from a list of rgb arrays
    """
    print("Rendering video...")
    if vid_path[-1] != '/':
        vid_path += '/'
    video_path = vid_path + video_name + '.mp4'

    if resize is not None:
        width, height = resize
    else:
        frame = rgb_arrs[0]
        height, width, layers = frame.shape

    fourcc = cv2.VideoWriter_fourcc(*format)
    video = cv2.VideoWriter(video_path, fourcc, float(fps), (width, height))

    for i, image in enumerate(rgb_arrs):
        percent_done = int((i / len(rgb_arrs)) * 100)
        if percent_done % 20 == 0:
            print("\t...", percent_done, "% of frames rendered")
        if resize is not None:
            image = cv2.resize(image, resize, interpolation=cv2.INTER_NEAREST)
        video.write(image)

    video.release()
    cv2.destroyAllWindows()


def return_view(grid, pos, row_size, col_size):
    """Given a map grid, position and view window, returns correct map part

    Note, if the agent asks for a view that exceeds the map bounds,
    it is padded with zeros

    Parameters
    ----------
    grid: 2D array
        map array containing characters representing
    pos: list
        list consisting of row and column at which to search
    row_size: int
        how far the view should look in the row dimension
    col_size: int
        how far the view should look in the col dimension

    Returns
    -------
    view: (np.ndarray) - a slice of the map for the agent to see
    """
    x, y = pos
    left_edge = x - col_size
    right_edge = x + col_size
    top_edge = y - row_size
    bot_edge = y + row_size
    pad_mat, left_pad, top_pad = pad_if_needed(left_edge, right_edge,
                                               top_edge, bot_edge, grid)
    x += left_pad
    y += top_pad
    view = pad_mat[x - col_size: x + col_size + 1,
                   y - row_size: y + row_size + 1]
    return view


def pad_if_needed(left_edge, right_edge, top_edge, bot_edge, matrix):
    # FIXME(ev) something is broken here, I think x and y are flipped
    row_dim = matrix.shape[0]
    col_dim = matrix.shape[1]
    left_pad, right_pad, top_pad, bot_pad = 0, 0, 0, 0
    if left_edge < 0:
        left_pad = abs(left_edge)
    if right_edge > row_dim - 1:
        right_pad = right_edge - (row_dim - 1)
    if top_edge < 0:
        top_pad = abs(top_edge)
    if bot_edge > col_dim - 1:
        bot_pad = bot_edge - (col_dim - 1)

    return pad_matrix(left_pad, right_pad, top_pad, bot_pad, matrix, 0), left_pad, top_pad


def pad_matrix(left_pad, right_pad, top_pad, bot_pad, matrix, const_val=1):
    pad_mat = np.pad(matrix, ((left_pad, right_pad), (top_pad, bot_pad)),
                     'constant', constant_values=(const_val, const_val))
    return pad_mat



# -*- coding: utf-8 -*-

import gym
from gym.utils import seeding
import numpy as np
from gym.spaces import Box, Tuple, MultiDiscrete
import random
import sys, os
from contextlib import closing
from io import StringIO
from gym import utils

# We need this for the config when calling from a different dir:
sys.path.append(os.path.dirname(__file__)) 
from environments.MultiTaxiEnv.config import TAXI_ENVIRONMENT_REWARDS, BASE_AVAILABLE_ACTIONS, ALL_ACTIONS_NAMES

orig_MAP = [
    "+---------+",
    "|X: |F: :X|",
    "| : | : : |",
    "| : : : : |",
    "| | : | : |",
    "|X| :G|X: |",
    "+---------+",
]

MAP2 = [
    "+-------+",
    "|X: |F:X|",
    "| : | : |",
    "| : : : |",
    "|X| :G|X|",
    "+-------+",
]

MAP = [
    "+-----------------------+",
    "|X: |F: | : | : | : |F:X|",
    "| : | : : : | : | : | : |",
    "| : : : : : : : : : : : |",
    "| : : : : : | : : : : : |",
    "| : : : : : | : : : : : |",
    "| : : : : : : : : : : : |",
    "|X| :G| | | :G| | | : |X|",
    "+-----------------------+",
]

simple_MAP = [
    "+---------+",
    "| : : : : |",
    "| : : : : |",
    "| : : : : |",
    "| : : : : |",
    "| : : : : |",
    "+---------+",
]


class TaxiEnv:
    """
    The Taxi Problem
    from "Hierarchical Reinforcement Learning with the MAXQ Value Function Decomposition"
    by Tom Dietterich
    Description:
    There are four designated locations in the grid world indicated by R(ed), G(reen), Y(ellow), and B(lue).
    When the episode starts, the taxi starts off at a random square and the passenger is at a random location.
    The taxi drives to the passenger's location, picks up the passenger, drives to the passenger's destination
    (another one of the four specified locations), and then drops off the passenger. Once the passenger is dropped off,
    the episode ends.

    Observations:
    A list (taxis, fuels, pass_start, destinations, pass_locs):
        taxis:                  a list of coordinates of each taxi
        fuels:                  a list of fuels for each taxi
        pass_start:             a list of starting coordinates for taeach passenger (current position or last available)
        destinations:           a list of destination coordinates for each passenger
        passengers_locations:   a list of locations of each passenger.
                                -1 means delivered
                                0 means not picked up
                                positive number means the passenger is in the corresponding taxi number

    Passenger start: coordinates of each of these
    - -1: In a taxi
    - 0: R(ed)
    - 1: G(reen)
    - 2: Y(ellow)
    - 3: B(lue)

    Passenger location:
    - -1: delivered
    - 0: not in taxi
    - x: in taxi x (x is integer)

    Destinations: coordinates of each of these
    - 0: R(ed)
    - 1: G(reen)
    - 2: Y(ellow)
    - 3: B(lue)

    Fuel:
     - 0 to np.inf: default with 10

    Actions:
    Actions are given as a list, each element referring to one taxi's action. Each taxi has 7 actions:
    - 0: move south
    - 1: move north
    - 2: move east
    - 3: move west
    - 4: pickup passenger
    - 5: dropoff passenger
    - 6: turn engine on
    - 7: turn engine off
    - 8: standby
    - 9: refuel fuel tank


    Rewards:
    - Those are specified in the config file.

    Rendering:
    - blue: passenger
    - magenta: destination
    - yellow: empty taxi
    - green: full taxi
    - other letters (R, G, Y and B): locations for passengers and destinations
    Main class to be characterized with hyper-parameters.
    """

    metadata = {'render.modes': ['human', 'ansi']}

    def __init__(self, _=0, num_taxis: int = 2, num_passengers: int = 1, max_fuel: list = None,
                 domain_map: list = None, taxis_capacity: list = None, collision_sensitive_domain: bool = True,
                 fuel_type_list: list = None, option_to_stand_by: bool = False):
        """
        TODO -  later version make number of passengers dynamic, even in runtime
        Args:
            num_taxis: number of taxis in the domain
            num_passengers: number of passengers occupying the domain at initiailaization
            max_fuel: list of max (start) fuel, we use np.inf as default for fuel free taxi.
            domain_map: 2D - map of the domain
            taxis_capacity: max capacity of passengers in each taxi (list)
            collision_sensitive_domain: is the domain show and react (true) to collisions or not (false)
            fuel_type_list: list of fuel types of each taxi
            option_to_stand_by: can taxis simply stand in place
        """
        # Initializing default values
        self.num_taxis = num_taxis
        if max_fuel is None:
            self.max_fuel = [100] * num_taxis  # TODO - needs to figure out how to insert np.inf into discrete obs.space
        else:
            self.max_fuel = max_fuel

        if domain_map is None:
            self.desc = np.asarray(orig_MAP, dtype='c')
        else:
            self.desc = np.asarray(domain_map, dtype='c')

        if taxis_capacity is None:
            self.taxis_capacity = [1] * num_taxis
        else:
            self.taxis_capacity = taxis_capacity

        if fuel_type_list is None:
            self.fuel_type_list = ['F'] * num_taxis
        else:
            self.fuel_type_list = fuel_type_list

        # Relevant features for map orientation, notice that we can only drive between the columns (':')
        self.num_rows = num_rows = len(self.desc) - 2
        self.num_columns = num_columns = len(self.desc[0][1:-1:2])

        # Set locations of passengers and fuel stations according to the map.
        self.passengers_locations = []
        self.fuel_station1 = None
        self.fuel_station2 = None
        self.fuel_stations = []

        # initializing map with passengers and fuel stations
        for i, row in enumerate(self.desc[1:-1]):
            for j, char in enumerate(row[1:-1:2]):
                loc = [i, j]
                if char == b'X':
                    self.passengers_locations.append(loc)
                elif char == b'F':
                    self.fuel_station1 = loc
                    self.fuel_stations.append(loc)
                elif char == b'G':
                    self.fuel_station2 = loc
                    self.fuel_stations.append(loc)

        self.coordinates = [[i, j] for i in range(num_rows) for j in range(num_columns)]

        # self.num_taxis = num_taxis
        self.taxis_names = ["taxi_" + str(index + 1) for index in range(self.num_taxis)]

        self.collision_sensitive_domain = collision_sensitive_domain

        # Indicator list of 1's (collided) and 0's (not-collided) of all taxis
        self.collided = np.zeros(self.num_taxis)

        self.option_to_standby = option_to_stand_by

        # A list to indicate whether the engine of taxi i is on (1) or off (0), all taxis start as on.
        self.engine_status_list = list(np.ones(self.num_taxis).astype(bool))

        self.num_passengers = num_passengers

        # Available actions in relation to all actions based on environment parameters.
        self.available_actions_indexes, self.index_action_dictionary, self.action_index_dictionary \
            = self._set_available_actions_dictionary()
        self.num_actions = len(self.available_actions_indexes)
        self.action_space = gym.spaces.Discrete(self.num_actions)
        self.observation_space = gym.spaces.MultiDiscrete(self._get_observation_space_list())
        self.bounded = False

        self.last_action = None
        self.num_states = self._get_num_states()

        self._seed()
        self.state = None
        self.dones = {taxi_name: False for taxi_name in self.taxis_names}
        self.dones['__all__'] = False

        self.np_random = None
        self.reset()

    def _get_num_states(self):
        map_dim = (self.num_rows * self.num_columns)
        passengers_loc_dim = 1
        for i in range(self.num_passengers):
            passengers_loc_dim *= len(self.passengers_locations) + self.num_taxis - i
        passengers_dest_dim = 1
        for i in range(self.num_passengers):
            passengers_dest_dim *= len(self.passengers_locations) - i
        num_states = map_dim * passengers_loc_dim * passengers_dest_dim
        return num_states

    def _get_observation_space_list(self) -> list:
        """
        Returns a list that emebed the observation space size in each dimension.
        An observation is a list of the form:
        [
            taxi_row, taxi_col, taxi_fuel,
            passenger1_row, passenger1_col,
            ...
            passenger_n_row, passenger_n_col,
            passenger1_dest_row, passenger1_dest_col,
            ...
            passenger_n_dest_row, passenger_n_dest_col,
            passenger1_status,
            ...
            passenger_n_status
        ]
        Returns: a list with all the dimensions sizes of the above.

        """
        locations_sizes = [self.num_rows, self.num_columns]
        fuel_size = [max(self.max_fuel) + 1]
        passengers_status_size = [self.num_taxis + 3]
        dimensions_sizes = []

        for _ in range(self.num_taxis):
            dimensions_sizes += locations_sizes
        for _ in range(self.num_taxis):
            dimensions_sizes += fuel_size

        for _ in range(self.num_passengers):
            dimensions_sizes += 2 * locations_sizes
        for _ in range(self.num_passengers):
            dimensions_sizes += passengers_status_size

        return [dimensions_sizes]

    def _seed(self, seed=None) -> list:
        """
        Setting a seed for the random sample state generation.
        Args:
            seed: seed to use

        Returns: list[seed]

        """
        self.np_random, self.seed_id = seeding.np_random(seed)
        return np.array([self.seed_id])

    def reset(self) -> dict:
        """
        Reset the environment's state:
            - taxis coordinates.
            - refuel all taxis
            - random get destinations.
            - random locate passengers.
            - preserve other definitions of the environment (collision, capacity...)
            - all engines turn on.
        Args:

        Returns: The reset state.

        """
        # reset taxis locations
        taxis_locations = random.sample(self.coordinates, self.num_taxis)
        self.collided = np.zeros(self.num_taxis)
        self.bounded = False
        self.window_size = 5
        self.counter = 0

        # refuel everybody
        fuels = [self.max_fuel[i] for i in range(self.num_taxis)]

        # reset passengers
        passengers_start_location = [start for start in
                                     random.choices(self.passengers_locations, k=self.num_passengers)]
        passengers_destinations = [random.choice([x for x in self.passengers_locations if x != start])
                                   for start in passengers_start_location]

        # Status of each passenger: delivered (1), in_taxi (positive number>2), waiting (2)
        passengers_status = [2 for _ in range(self.num_passengers)]
        self.state = [taxis_locations, fuels, passengers_start_location, passengers_destinations, passengers_status]

        self.last_action = None
        # Turning all engines on
        self.engine_status_list = list(np.ones(self.num_taxis))

        # resetting dones
        self.dones = {taxi_id: False for taxi_id in self.taxis_names}
        self.dones['__all__'] = False
        obs = {}
        for taxi_id in self.taxis_names:
            obs[taxi_id] = self.get_observation(self.state, taxi_id)

        return obs

    def _set_available_actions_dictionary(self) -> (list, dict, dict):
        """

        TODO: Later versions - maybe return an action-dictionary for each taxi individually.

        Generates list of all available actions in the parametrized domain, index->action dictionary to decode.
        Generation is based on the hyper-parameters passed to __init__ + parameters defined in config.py

        Returns: list of available actions, index->action dictionary for all actions and the reversed dictionary
        (action -> index).

        """

        action_names = ALL_ACTIONS_NAMES  # From config.py
        base_dictionary = {}  # Total dictionary{index -> action_name}
        for index, action in enumerate(action_names):
            base_dictionary[index] = action

        available_action_list = BASE_AVAILABLE_ACTIONS  # From config.py

        if self.option_to_standby:
            available_action_list += ['turn_engine_on', 'turn_engine_off', 'standby']

        # TODO - when we return dictionary per taxi we can't longer assume that on np.inf fuel
        #  means no limited fuel for all the taxis
        if not self.max_fuel[0] == np.inf:
            available_action_list.append('refuel')

        action_index_dictionary = dict((value, key) for key, value in base_dictionary.items())  # {action -> index} all
        available_actions_indexes = [action_index_dictionary[action] for action in available_action_list]
        index_action_dictionary = dict((key, value) for key, value in base_dictionary.items())

        return list(set(available_actions_indexes)), index_action_dictionary, action_index_dictionary

    def get_available_actions_dictionary(self) -> (list, dict):
        """
        Returns: list of available actions and index->action dictionary for all actions.

        """
        return self.available_actions_indexes, self.index_action_dictionary

    def _is_there_place_on_taxi(self, passengers_locations: np.array, taxi_index: int) -> bool:
        """
        Checks if there is room for another passenger on taxi number 'taxi_index'.
        Args:
            passengers_locations: list of all passengers locations
            taxi_index: index of the desired taxi

        Returns: Whether there is a place (True) or not (False)

        """
        # Remember that passengers "location" is: 1 - delivered, 2 - waits for a taxi, >2 - on a taxi with index
        # location+2

        return (len([location for location in passengers_locations if location == (taxi_index + 3)]) <
                self.taxis_capacity[taxi_index])

    def map_at_location(self, location: list) -> str:
        """
        Returns the map character on the specified coordinates of the grid.
        Args:
            location: location to check [row, col]

        Returns: character on specific location on the map

        """
        domain_map = self.desc.copy().tolist()
        row, col = location[0], location[1]
        return domain_map[row + 1][2 * col + 1].decode(encoding='UTF-8')

    def at_valid_fuel_station(self, taxi: int, taxis_locations: list) -> bool:
        """
        Checks if the taxi's location is a suitable fuel station or not.
        Args:
            taxi: the index of the desired taxi
            taxis_locations: list of taxis coordinates [row, col]
        Returns: whether the taxi is at a suitable fuel station (true) or not (false)

        """
        return (taxis_locations[taxi] in self.fuel_stations and
                self.map_at_location(taxis_locations[taxi]) == self.fuel_type_list[taxi])

    def _get_action_list(self, action_list) -> list:
        """
        Return a list in the correct format for the step function that should
        always get a list even if it's a single action.
        Args:
            action_list:

        Returns: list(action_list)

        """
        if type(action_list) != list:
            return [action_list]

        return action_list

    def _engine_is_off_actions(self, action: str, taxi: int) -> int:
        """
        Returns the reward according to the requested action given that the engine's is currently off.
        Also turns engine on if requested.
        Args:
            action: requested action
            taxi: index of the taxi specified, relevant for turning engine on
        Returns: correct reward

        """
        reward = self.partial_closest_path_reward('unrelated_action')
        if action == 'standby':  # standby while engine is off
            reward = self.partial_closest_path_reward('standby_engine_off')
        elif action == 'turn_engine_on':  # turn engine on
            reward = self.partial_closest_path_reward('turn_engine_on')
            self.engine_status_list[taxi] = 1

        return reward

    def _take_movement(self, action: str, row: int, col: int) -> (bool, int, int):
        """
        Takes a movement with regard to a apecific location of a taxi,
        Args:
            action: direction to move
            row: current row
            col: current col

        Returns: if moved (false if there is a wall), new row, new col

        """
        moved = False
        new_row, new_col = row, col
        max_row = self.num_rows - 1
        max_col = self.num_columns - 1
        if action == 'south':  # south
            if row != max_row:
                moved = True
            new_row = min(row + 1, max_row)
        elif action == 'north':  # north
            if row != 0:
                moved = True
            new_row = max(row - 1, 0)
        if action == 'east' and self.desc[1 + row, 2 * col + 2] == b":":  # east
            if col != max_col:
                moved = True
            new_col = min(col + 1, max_col)
        elif action == 'west' and self.desc[1 + row, 2 * col] == b":":  # west
            if col != 0:
                moved = True
            new_col = max(col - 1, 0)

        return moved, new_row, new_col

    def _check_action_for_collision(self, taxi_index: int, taxis_locations: list, current_row: int, current_col: int,
                                    moved: bool, current_action: int, current_reward: int) -> (int, bool, int, list):
        """
        Takes a desired location for a taxi and update it with regard to collision check.
        Args:
            taxi_index: index of the taxi
            taxis_locations: locations of all other taxis.
            current_row: of the taxi
            current_col: of the taxi
            moved: indicator variable
            current_action: the current action requested
            current_reward: the current reward (left unchanged if there is no collision)

        Returns: new_reward, new_moved, new_action_index

        """
        reward = current_reward
        row, col = current_row, current_col
        moved = moved
        action = current_action
        taxi = taxi_index
        # Check if the number of taxis on the destination location is greater than 0
        if len([i for i in range(self.num_taxis) if taxis_locations[i] == [row, col]]) > 0:
            if self.option_to_standby:
                moved = False
                action = self.action_index_dictionary['standby']
            else:
                self.collided[[i for i in range(len(taxis_locations)) if taxis_locations[i] == [row, col]]] = 1
                self.collided[taxi] = 1
                reward = self.partial_closest_path_reward('collision')
                taxis_locations[taxi] = [row, col]

        return reward, moved, action, taxis_locations

    def _make_pickup(self, taxi: int, passengers_start_locations: list, passengers_status: list,
                     taxi_location: list, reward: int) -> (list, int):
        """
        Make a pickup (successful or fail) for a given taxi.
        Args:
            taxi: index of the taxi
            passengers_start_locations: current locations of the passengers
            passengers_status: list of passengers statuses (1, 2, greater..)
            taxi_location: location of the taxi
            reward: current reward

        Returns: updates passengers status list, updates reward

        """
        passengers_status = passengers_status
        reward = reward
        successful_pickup = False
        for i, location in enumerate(passengers_status):
            # Check if we can take this passenger
            if location == 2 and taxi_location == passengers_start_locations[i] and \
                    self._is_there_place_on_taxi(passengers_status, taxi):
                passengers_status[i] = taxi + 3
                successful_pickup = True
                reward = self.partial_closest_path_reward('pickup')
        if not successful_pickup:  # passenger not at location
            reward = self.partial_closest_path_reward('bad_pickup')

        return passengers_status, reward

    def _make_dropoff(self, taxi: int, current_passengers_start_locations: list, current_passengers_status: list,
                      destinations: list, taxi_location: list, reward: int) -> (list, list, int):
        """
        Make a dropoff (successful or fail) for a given taxi.
        Args:
            taxi: index of the taxi
            current_passengers_start_locations: current locations of the passengers
            current_passengers_status: list of passengers statuses (1, 2, greater..)
            destinations: list of passengers destinations
            taxi_location: location of the taxi
            reward: current reward

        Returns: updates passengers status list, updated passengers start location, updates reward

        """
        reward = reward
        passengers_start_locations = current_passengers_start_locations.copy()
        passengers_status = current_passengers_status.copy()
        successful_dropoff = False
        for i, location in enumerate(passengers_status):  # at destination
            location = passengers_status[i]
            # Check if we have the passenger and we are at his destination
            if location == (taxi + 3) and taxi_location == destinations[i]:
                passengers_status[i] = 1
                reward = self.partial_closest_path_reward('final_dropoff', taxi)
                passengers_start_locations[i] = taxi_location
                successful_dropoff = True
                break
            elif location == (taxi + 3):  # drops off passenger not at destination
                passengers_status[i] = 2
                successful_dropoff = True
                reward = self.partial_closest_path_reward('intermediate_dropoff', taxi)
                passengers_start_locations[i] = taxi_location
                break
        if not successful_dropoff:  # not carrying a passenger
            reward = self.partial_closest_path_reward('bad_dropoff')

        return passengers_status, passengers_start_locations, reward

    def _update_movement_wrt_fuel(self, taxi: int, taxis_locations: list, wanted_row: int, wanted_col: int,
                                  reward: int, fuel: int) -> (int, int, list):
        """
        Given that a taxi would like to move - check the fuel accordingly and update reward and location.
        Args:
            taxi: index of the taxi
            taxis_locations: list of current locations (prior to movement)
            wanted_row: row after movement
            wanted_col: col after movement
            reward: current reward
            fuel: current fuel

        Returns: updated_reward, updated fuel, updared_taxis_locations

        """
        reward = reward
        fuel = fuel
        taxis_locations = taxis_locations
        if fuel == 0:
            reward = ('no_fuel')
        else:
            fuel = max(0, fuel - 1)

            taxis_locations[taxi] = [wanted_row, wanted_col]

        return reward, fuel, taxis_locations

    def _refuel_taxi(self, current_fuel: int, current_reward: int, taxi: int, taxis_locations: list) -> (int, int):
        """
        Try to refuel a taxi, if successful - updates fuel tank, if not - updates the reward.
        Args:
            current_fuel: current fuel of the taxi
            current_reward: current reward for the taxi.
            taxi: taxi index
            taxis_locations: list of current taxis locations

        Returns: updated reward, updated fuel

        """
        fuel = current_fuel
        reward = current_reward
        if self.at_valid_fuel_station(taxi, taxis_locations) and fuel != self.max_fuel[taxi]:
            fuel = self.max_fuel[taxi]
        else:
            reward = self.partial_closest_path_reward('bad_refuel')

        return reward, fuel

    def step(self, action_dict: dict) -> (dict, dict, dict, dict):
        """
        Executing a list of actions (action for each taxi) at the domain current state.
        Supports not-joined actions, just pass 1 element instead of list.

        Args:
            action_dict: {taxi_name: action} - action of specific taxis to take on the step

        Returns: - dict{taxi_id: observation}, dict{taxi_id: reward}, dict{taxi_id: done}, _
        """

        rewards = {}
        self.counter += 1
        if self.counter >= 90:
            self.window_size = 3

        # Main of the function, for each taxi-i act on action[i]
        for taxi_name, action_list in action_dict.items():
            # meta operations on the type of the action
            action_list = self._get_action_list(action_list)

            for action in action_list:
                taxi = self.taxis_names.index(taxi_name)
                reward = self.partial_closest_path_reward('step')  # Default reward
                moved = False  # Indicator variable for later use

                # taxi locations: [i, j]
                # fuels: int
                # passengers_start_locations and destinations: [[i, j] ... [i, j]]
                # passengers_status: [[1, 2, taxi_index+2] ... [1, 2, taxi_index+2]], 1 - delivered
                taxis_locations, fuels, passengers_start_locations, destinations, passengers_status = self.state

                if all(list(self.dones.values())):
                    rewards[taxi_name] = reward
                    continue

                # If taxi is collided, it can't perform a step
                if self.collided[taxi] == 1:
                    rewards[taxi_name] = self.partial_closest_path_reward('collided')
                    self.dones[taxi_name] = True
                    continue

                # If the taxi is out of fuel, it can't perform a step
                if fuels[taxi] == 0 and not self.at_valid_fuel_station(taxi, taxis_locations):
                    rewards[taxi_name] = self.partial_closest_path_reward('bad_fuel')
                    self.dones[taxi_name] = True
                    continue

                taxi_location = taxis_locations[taxi]
                row, col = taxi_location

                fuel = fuels[taxi]
                is_taxi_engine_on = self.engine_status_list[taxi]
                _, index_action_dictionary = self.get_available_actions_dictionary()

                if not is_taxi_engine_on:  # Engine is off
                    # update reward according to standby/ turn-on/ unrelated + turn engine on if requsted
                    reward = self._engine_is_off_actions(index_action_dictionary[action], taxi)


                else:  # Engine is on
                    # Binding
                    if index_action_dictionary[action] == 'bind':
                        self.bounded = False
                        reward = self.partial_closest_path_reward('bind')

                    # Movement
                    if index_action_dictionary[action] in ['south', 'north', 'east', 'west']:
                        moved, row, col = self._take_movement(index_action_dictionary[action], row, col)

                    # Check for collisions
                    if self.collision_sensitive_domain and moved:
                        if self.collided[taxi] == 0:
                            reward, moved, action, taxis_locations = self._check_action_for_collision(taxi,
                                                                                                      taxis_locations,
                                                                                                      row, col, moved,
                                                                                                      action, reward)

                    # Pickup
                    elif index_action_dictionary[action] == 'pickup':
                        passengers_status, reward = self._make_pickup(taxi, passengers_start_locations,
                                                                      passengers_status, taxi_location, reward)

                    # Dropoff
                    elif index_action_dictionary[action] == 'dropoff':
                        passengers_status, passengers_start_locations, reward = self._make_dropoff(taxi,
                                                                                                   passengers_start_locations,
                                                                                                   passengers_status,
                                                                                                   destinations,
                                                                                                   taxi_location,
                                                                                                   reward)

                    # Turning engine off
                    elif index_action_dictionary[action] == 'turn_engine_off':
                        reward = self.partial_closest_path_reward('turn_engine_off')
                        self.engine_status_list[taxi] = 0

                    # Standby with engine on
                    elif index_action_dictionary[action] == 'standby':
                        reward = self.partial_closest_path_reward('standby_engine_on')

                # Here we have finished checking for action for taxi-i
                # Fuel consumption
                if moved:
                    reward, fuels[taxi], taxis_locations = self._update_movement_wrt_fuel(taxi, taxis_locations,
                                                                                          row, col, reward, fuel)

                if (not moved) and action in [self.action_index_dictionary[direction] for
                                              direction in ['north', 'south', 'west', 'east']]:
                    reward = TAXI_ENVIRONMENT_REWARDS['hit_wall']

                # taxi refuel
                if index_action_dictionary[action] == 'refuel':
                    reward, fuels[taxi] = self._refuel_taxi(fuel, reward, taxi, taxis_locations)

                # check if all the passengers are at their destinations
                done = all(loc == 1 for loc in passengers_status)
                self.dones[taxi_name] = done

                # check if all taxis collided
                done = all(self.collided == 1)
                self.dones[taxi_name] = self.dones[taxi_name] or done

                # check if all taxis are out of fuel
                done = fuels[taxi] == 0
                self.dones[taxi_name] = self.dones[taxi_name] or done

                rewards[taxi_name] = reward
                self.state = [taxis_locations, fuels, passengers_start_locations, destinations, passengers_status]
                self.last_action = action_dict

        self.dones['__all__'] = True
        self.dones['__all__'] = all(list(self.dones.values()))

        if self.bounded:
            total_reward = 0
            for taxi_id in action_dict.keys():
                total_reward += rewards[taxi_id]
            total_reward /= len(action_dict.keys())
            for taxi_id in action_dict.keys():
                rewards[taxi_id] = total_reward

        obs = {}
        for taxi_id in action_dict.keys():
            obs[taxi_id] = self.get_observation(self.state, taxi_id)

        return obs, {taxi_id: rewards[taxi_id] for taxi_id in action_dict.keys()}, self.dones, {}

    def render(self, mode: str = 'human') -> str:
        """
        Renders the domain map at the current state
        Args:
            mode: Demand mode (file or human watching).

        Returns: Value string of writing the output

        """
        outfile = StringIO() if mode == 'ansi' else sys.stdout

        # Copy map to work on
        out = self.desc.copy().tolist()
        out = [[c.decode('utf-8') for c in line] for line in out]

        taxis, fuels, passengers_start_coordinates, destinations, passengers_locations = self.state

        colors = ['yellow', 'red', 'white', 'green', 'cyan', 'crimson', 'gray', 'magenta'] * 5
        colored = [False] * self.num_taxis

        def ul(x):
            """returns underline instead of spaces when called"""
            return "_" if x == " " else x

        for i, location in enumerate(passengers_locations):
            if location > 2:  # Passenger is on a taxi
                taxi_row, taxi_col = taxis[location - 3]

                # Coloring taxi's coordinate on the map
                out[1 + taxi_row][2 * taxi_col + 1] = utils.colorize(
                    out[1 + taxi_row][2 * taxi_col + 1], colors[location - 3], highlight=True, bold=True)
                colored[location - 3] = True
            else:  # Passenger isn't in a taxi
                # Coloring passenger's coordinates on the map
                pi, pj = passengers_start_coordinates[i]
                out[1 + pi][2 * pj + 1] = utils.colorize(out[1 + pi][2 * pj + 1], 'blue', bold=True)

        for i, taxi in enumerate(taxis):
            if self.collided[i] == 0:  # Taxi isn't collided
                taxi_row, taxi_col = taxi
                out[1 + taxi_row][2 * taxi_col + 1] = utils.colorize(
                    ul(out[1 + taxi_row][2 * taxi_col + 1]), colors[i], highlight=True)
            else:  # Collided!
                taxi_row, taxi_col = taxi
                out[1 + taxi_row][2 * taxi_col + 1] = utils.colorize(
                    ul(out[1 + taxi_row][2 * taxi_col + 1]), 'gray', highlight=True)

        for dest in destinations:
            di, dj = dest
            out[1 + di][2 * dj + 1] = utils.colorize(out[1 + di][2 * dj + 1], 'magenta')
        outfile.write("\n".join(["".join(row) for row in out]) + "\n")

        if self.last_action is not None:
            moves = ALL_ACTIONS_NAMES
            output = [moves[i] for i in np.array(list(self.last_action.values())).reshape(-1)]
            outfile.write("  ({})\n".format(' ,'.join(output)))
        for i, taxi in enumerate(taxis):
            outfile.write("Taxi{}-{}: Fuel: {}, Location: ({},{}), Collided: {}\n".format(i + 1, colors[i].upper(),
                                                                                          fuels[i], taxi[0], taxi[1],
                                                                                          self.collided[i] == 1))
        for i, location in enumerate(passengers_locations):
            start = tuple(passengers_start_coordinates[i])
            end = tuple(destinations[i])
            if location == 1:
                outfile.write("Passenger{}: Location: Arrived!, Destination: {}\n".format(i + 1, end))
            if location == 2:
                outfile.write("Passenger{}: Location: {}, Destination: {}\n".format(i + 1, start, end))
            else:
                outfile.write("Passenger{}: Location: Taxi{}, Destination: {}\n".format(i + 1, location - 2, end))
        outfile.write("Done: {}, {}\n".format(all(self.dones.values()), self.dones))
        outfile.write("Passengers Status's: {}\n".format(self.state[-1]))

        # No need to return anything for human
        if mode != 'human':
            with closing(outfile):
                return outfile.getvalue()

    @staticmethod
    def partial_observations(state: list) -> list:
        """
        Get partial observation of state.
        Args:
            state: state of the domain (taxis, fuels, passengers_start_coordinates, destinations, passengers_locations)

        Returns: list of observations s.t each taxi sees only itself

        """

        def flatten(x):
            return [item for sub in x for item in sub]

        observations = []
        taxis, fuels, passengers_start_locations, passengers_destinations, passengers_locations = state
        pass_info = flatten(passengers_start_locations) + flatten(passengers_destinations) + passengers_locations

        for i in range(len(taxis)):
            obs = taxis[i] + [fuels[i]] + pass_info
            obs = np.reshape(obs, [1, len(obs)])
            observations.append(obs)
        return observations

    def get_l1_distance(self, location1, location2):
        """
        Return the minimal travel length between 2 locations on the grid world.
        Args:
            location1: [i1, j1]
            location2: [i2, j2]

        Returns: np.abs(i1 - i2) + np.abs(j1 - j2)

        """
        return np.abs(location1[0] - location2[0]) + np.abs(location1[1] - location2[1])

    def get_observation(self, state: list, agent_name: str) -> np.array:
        """
        Takes only the observation of the specified agent.
        Args:
            state: state of the domain (taxis, fuels, passengers_start_coordinates, destinations, passengers_locations)
            agent_name: observer name
            window_size: the size that the agent can see in the map (around it) in terms of other txis

        Returns: observation of the specified agent (state wise)

        """

        def flatten(x):
            return [item for sub in list(x) for item in list(sub)]

        if type(agent_name) == str:
            agent_index = self.taxis_names.index(agent_name)
        else:
            agent_index = agent_name

        taxis, fuels, passengers_start_locations, passengers_destinations, passengers_locations = state.copy()
        passengers_information = flatten(passengers_start_locations) + flatten(
            passengers_destinations) + passengers_locations

        closest_taxis_indices = []
        for i in range(self.num_taxis):
            if self.get_l1_distance(taxis[agent_index], taxis[i]) <= self.window_size and i != agent_index:
                closest_taxis_indices.append(i)

        observations = taxis[agent_index].copy()
        for i in closest_taxis_indices:
            observations += taxis[i]
        observations += [0, 0] * (self.num_taxis - 1 - len(closest_taxis_indices)) + [fuels[agent_index]] + \
                        [0] * (self.num_taxis - 1) + passengers_information
        observations = np.reshape(observations, (1, len(observations)))

        return observations

    def passenger_destination_l1_distance(self, passenger_index, current_row: int, current_col: int) -> int:
        """
        Returns the manhattan distance between passenger current defined "start location" and it's destination.
        Args:
            passenger_index: index of the passenger.
            current_row: current row to calculate distance from destination
            current_col: current col to calculate distance from destination

        Returns: manhattan distance

        """
        current_state = self.state
        destination_row, destination_col = current_state[3][passenger_index]
        return int(np.abs(current_col - destination_col) + np.abs(current_row - destination_row))

    def partial_closest_path_reward(self, basic_reward_str: str, taxi_index: int = None) -> int:
        """
        Computes the reward for a taxi and it's defined by:
        dropoff[s] - gets the reward equal to the closest path multiply by 15, if the drive got a passenger further
        away - negative.
        other actions - basic reward from config table
        Args:
            basic_reward_str: the reward we would like to give
            taxi_index: index of the specific taxi

        Returns: updated reward

        """
        if basic_reward_str not in ['intermediate_dropoff', 'final_dropoff'] or taxi_index is None:
            return TAXI_ENVIRONMENT_REWARDS[basic_reward_str]

        # [taxis_locations, fuels, passengers_start_locations, destinations, passengers_status]
        current_state = self.state
        passengers_start_locations = current_state[2]

        taxis_locations = current_state[0]

        passengers_status = current_state[-1]
        passenger_index = passengers_status.index(taxi_index + 3)
        passenger_start_row, passenger_start_col = passengers_start_locations[passenger_index]
        taxi_current_row, taxi_current_col = taxis_locations[taxi_index]

        return 15 * (self.passenger_destination_l1_distance(passenger_index, passenger_start_row, passenger_start_col) -
                     self.passenger_destination_l1_distance(passenger_index, taxi_current_row, taxi_current_col))


from abc import abstractmethod
from control.controller import Controller
import numpy as np

"""Abstract parent class for centralized controller 
"""
class Centralized(Controller):

    def __init__(self, env, agents, central_agent):
        # initialize super class
        super().__init__(env, agents, central_agent)
        self.decision_maker = self.central_agent.get_decision_maker()

    @abstractmethod
    def get_joint_action(self, observation):
        """Returns the joint action from all the agent

        Args:
            observation (dict): All the observations
        """
        pass

    @abstractmethod
    def decode_state(self, obs):
        """Abstract method - translate from the env observations
        into an ndarray for the model process
        """
        pass

    @abstractmethod
    def decode_action(self, action, num_actions, num_agents):
        """Abstract methods - translate from the model action
        into individual actions for every agent
        """
        pass


import numpy as np
from gym import spaces
from itertools import chain
from enum import IntEnum

BOARD_SIZE = 5
CORNER_REWARD = 100
CENTER_REWARD = 10
STEP_REWARD = -1

BOARD_UP =   "╔═══════════╗"
BOARD_LINE = "║           ║"
BOARD_DOWN = "╚═══════════╝"

class Action(IntEnum):noop=0;up=1;right=2;down=3;left=4

"""A simple environment for a Social Dilemma - 
We need one agent in every corner and one in the
middle to finish the episode - but the agent in the
middle gets less reward then the ones in the corners.
"""
class CornersEnv:

    def __init__(self):
        self.agents = ['A', 'B', 'C', 'D', 'E']
        
        self.action_spaces = {agent: spaces.Discrete(5) for agent in self.agents}

        # Observation: the pos of all other agents (4 agent, 2 coordinated each)
        self.observation_spaces = {agent: spaces.MultiDiscrete([5]*8) for agent in self.agents}
 
    def reset(self):
        # Random Start:
        self.pos = {}
        for agent in self.agents:
            new_pos = np.random.choice(range(BOARD_SIZE), size=(2,))
            while any([new_pos[0] == pos[0] and new_pos[1] == pos[1] for pos in self.pos.values()]):
                # If the random pos already exist just make a new one:
                new_pos = np.random.choice(range(BOARD_SIZE), size=(2,))
            self.pos[agent] = new_pos
        return self._make_observation()
        
    def render(self):
        lines = [list(BOARD_LINE) + ['\n'] for i in range(5)]
        for agent in self.agents:
            pos = self.pos[agent]
            lines[pos[0]][2 + 2*pos[1]] = agent
        board_list = list(BOARD_UP) + ['\n'] + list(chain.from_iterable(lines)) + list(BOARD_DOWN) + ['\n']
        print("".join(board_list))
        
    def step(self, actions):
        for agent in self.agents:
            if actions[agent] == Action.noop:
                continue
            elif actions[agent] == Action.up:
                new_pos = self.pos[agent] + [-1, 0]
            elif actions[agent] == Action.right:
                new_pos = self.pos[agent] + [0, 1]
            elif actions[agent] == Action.down:
                new_pos = self.pos[agent] + [1, 0]
            elif actions[agent] == Action.left:
                new_pos = self.pos[agent] + [0, -1]

            if self._is_legal_pos(new_pos):
                self.pos[agent] = new_pos

        if self._is_episode_over():
            rewards = {agent: CORNER_REWARD for agent in self.agents}
            dones = {agent: True for agent in self.agents}
        else:
            rewards = {agent: STEP_REWARD for agent in self.agents}
            dones = {agent: False for agent in self.agents}

        info = {}
        observations = self._make_observation()
        return observations, rewards, dones, info


    def _is_legal_pos(self, pos):
        already_occupied = any([pos[0] == p[0] and pos[1] == p[1] for p in self.pos.values()])
        return not already_occupied and (pos < BOARD_SIZE).all() and (pos >= 0).all()

    def _is_episode_over(self):
        target_pos = [
            [0,0], [4, 0], [0, 4], [4, 4], [2, 2]
        ]
        return all([
            any([pos[0] == target[0] and pos[1] == target[1] for pos in self.pos.values()])
        for target in target_pos])

    def _make_observation(self):
        return {
            agent: np.array(
                # The first element is the position of the current agent, then the rest of them
                [self.pos[agent]] + [self.pos[agent2] for agent2 in self.agents if agent2 != agent]
                )
        for agent in self.agents
        }


import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torch import autograd
from agents.agent import DecisionMaker


class DeepPolicyGradient(DecisionMaker):
    def __init__(self, input_dims, num_actions, is_conv=False, gamma=0.99, learning_rate=0.01,
                 mapping_fn=None):
        super().__init__()
        self.input_dims = input_dims[::-1]
        self.gamma = gamma
        self.reward_mem = []
        self.action_mem = []
        self.num_actions = num_actions

        if not is_conv:
            self.policy = PolicyNetwork(self.input_dims, num_actions, mapping_fn)
        else:
            self.policy = ConvPolicyNetwork(self.input_dims, num_actions)

        self.optimizer = optim.Adam(self.policy.parameters(), lr=learning_rate)

    def learn(self):
        """Updating the netwotk based on the memory"""
        self.optimizer.zero_grad()
        G = np.zeros_like(self.reward_mem, dtype=np.float64)

        for t in range(len(self.reward_mem)):
            G_sum = 0
            discount = 1

            for k in range(t, len(self.reward_mem)):
                G_sum += self.reward_mem[k] * discount
                discount *= self.gamma
            G[t] = G_sum

        mean = np.mean(G)
        std = np.std(G) if np.std(G) > 0 else 1
        G = (G - mean) / std
        G = torch.tensor(G, dtype=torch.float).to(self.policy.device)
        loss = 0

        for g, logprob in zip(G, self.action_mem):
            loss += -g * logprob

        loss.backward()
        self.optimizer.step()

    """ Training Callbacks """

    def get_train_action(self, observation):
        """Get action for when training"""
        probs = F.softmax(self.policy.forward(observation))
        action_probs = torch.distributions.Categorical(probs)
        action = action_probs.sample()
        log_probs = action_probs.log_prob(action)
        self.action_mem.append(log_probs)
        returned_action = action.item()
        return returned_action

    def update_step(self, obs, action, new_obs, reward, done):
        self.reward_mem.append(reward)

    def update_episode(self):
        self.learn()
        self.reward_mem = []
        self.action_mem = []

    """ Evaluation Callbacks """

    def get_action(self, observation):
        probs = F.softmax(self.policy.forward(observation))
        action_probs = torch.distributions.Categorical(probs)
        action = action_probs.sample()
        return action.item()


"""A class representing a fully connected network for the policy"""
class PolicyNetwork(nn.Module):

    def __init__(self, input_dims, output_dims, mapping_fn=None):
        super(PolicyNetwork, self).__init__()

        self.input_dim = input_dims
        self.output_dim = output_dims
        self.mapping_fn = mapping_fn
        self.device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')

        self.fc = nn.Sequential(
            nn.Linear(self.input_dim[0], 128),
            nn.ReLU(),
            nn.Linear(128, 256),
            nn.ReLU(),
            nn.Linear(256, self.output_dim)
        )

    def forward(self, observation):
        if self.mapping_fn:
            observation = self.mapping_fn(observation)

        state = torch.Tensor(observation).to(self.device)
        x = self.fc(state)
        return x


"""A class representing a fully convolution network for the policy"""
class ConvPolicyNetwork(nn.Module):

    def __init__(self, input_dims, output_dims):
        super(ConvPolicyNetwork, self).__init__()
        self.input_dim = input_dims
        self.output_dim = output_dims
        self.device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')

        self.conv = nn.Sequential(
            nn.Conv2d(self.input_dim[0], 32, kernel_size=8, stride=1),
            nn.ReLU(),
            nn.Conv2d(32, 64, kernel_size=4, stride=2),
            nn.ReLU(),
            nn.Conv2d(64, 64, kernel_size=3, stride=1),
            nn.ReLU()
        )

        self.fc_input_dim = self.feature_size()

        self.fc = nn.Sequential(
            nn.Linear(self.fc_input_dim, 128),
            nn.ReLU(),
            nn.Linear(128, 256),
            nn.ReLU(),
            nn.Linear(256, self.output_dim)
        )

    def forward(self, observation):
        observation = torch.Tensor(observation.copy()).unsqueeze(0).permute(0,3,1,2).to(self.device)
        features = self.conv(observation)
        features = features.view(features.size(0), -1)
        qvals = self.fc(features)
        return qvals

    def feature_size(self):
        return self.conv(autograd.Variable(torch.zeros(1, *self.input_dim))).view(1, -1).size(1)





from collections import defaultdict

import numpy as np

class EconomicControl:

    def __init__(self, env, agents, tasks):
        self.env = env
        self.agents = agents
        self.tasks = tasks
        self.given_tasks = defaultdict(list)

    def get_joint_action(self, observation):
        joint_action = {
            agent_name: self.agents[agent_name].get_action(observation[agent_name])
            for agent_name in self.agents.keys()     
        }
        return joint_action

    def run(self, max_episode_lenth=np.inf):
        """Runs the economic control on the env given in init

        Args:
            max_episode_lenth (int, optional): number of steps. Defaults to np.inf.
        """
        observation = self.env.reset()
        # We need to give the agents their position so they will know to bid
        for agent_name in self.agents.keys():
            self.agents[agent_name].reset(observation[agent_name])
        
        # Auction the tasks:
        self._auction_all_tasks()
        i = 1
        done = {'temp_agent': False}
        while(not all(value == True for value in done.values()) 
                and i < max_episode_lenth):
            print(f"Step {i}:")
            i += 1
            actions = self.get_joint_action(observation)
            observation, reward, done, info = self.env.step(actions)
            self.env.render()
        print(f"Finished")

    def _auction_all_tasks(self):
        """Auctions all the tasks given at init
        """
        non_busy_agents = list(self.agents.keys())
        for task in self.tasks:
            agent_chosen = self._auction_task(task, non_busy_agents)
            del(non_busy_agents[non_busy_agents.index(agent_chosen)])

    def _auction_task(self, task, agents):
        """Auction out tasks, and gives it to the agent that bids lowest.
        We remember the task so we can reward the agent when it is complete.
        """
        bids = {agent_name: self.agents[agent_name].get_bid(task) for agent_name in agents}
        agent_name = min(bids.keys(), key=(lambda k: bids[k]))
        self.agents[agent_name].set_task(task)
        self.given_tasks[agent_name].append(task)
        return agent_name


import numpy as np
from agents.agent import DecisionMaker


"""A PG algorithm for RL
"""
class PolicyGradient(DecisionMaker):

    def __init__(self, num_actions, num_obs, is_conv=False, num_agents=1, alpha=0.00025, gamma=0.9,
                 mapping_fn=None):
        super().__init__()
        self.num_obs = num_obs[::-1]
        self.alpha = alpha
        self.gamma = gamma

        """ Record keeping / episode """
        self.grads = []
        self.rewards = []

        self.mapping_fn = mapping_fn
        self.num_actions = num_actions
        self.num_agents = num_agents
        self.theta = np.random.rand(self.num_obs[0], self.num_actions)


    def softmax(self, state):
        z = state.dot(self.theta)
        exp = np.exp(z)
        return exp / np.sum(exp)

    def policy(self, state):
        probs = self.softmax(state)
        return probs

    def softmax_gradient(self, softmax):
        s = softmax.reshape(-1, 1)
        return np.diagflat(s) - np.dot(s, s.T)

    def compute_gradient(self, probs, state, action):
        dsoftmax_comp = self.softmax_gradient(probs)
        dsoftmax = dsoftmax_comp[action, :]
        dlog = dsoftmax / probs[0, action]
        grad = state.T.dot(dlog[None, :])

        self.grads.append(grad)
        return

    def update_weights(self):
        for i in range(len(self.grads)):
            present_val_of_rewards = sum([r * (self.gamma ** (t - i)) \
                                          for t, r in enumerate(self.rewards[i:])])
            self.theta += self.alpha * self.grads[i] * present_val_of_rewards * (self.gamma ** i)
        return

    """ Training Callbacks """

    def get_train_action(self, state):
        """Action for when training
        """
        if self.mapping_fn:
            state = self.mapping_fn(state)

        if self.mapping_fn:
            state = self.mapping_fn(state)

        state = state[None, :]
        probs = self.policy(state)
        action = np.random.choice(self.num_actions, p=probs[0])
        self.compute_gradient(probs, state, action)
        return action

    def update_step(self, obs, action, new_obs, reward, done):
        self.rewards.append(reward)

    def update_episode(self):
        """ Update weights, reset records for new episodes"""
        self.update_weights()

        self.grads = []
        self.rewards = []

    """ Evaluation callbacks """

    def get_action(self, state):
        """ Take an action according to policy."""
        if self.mapping_fn:
            state = self.mapping_fn(state)

        state = state[None, :]
        probs = self.policy(state)
        action = np.random.choice(self.num_actions, p=probs[0])
        return action



from collections import deque
import random
import numpy as np
import torch as torch
import torch.nn as nn
from torch import autograd
from agents.agent import DecisionMaker



"""A Deep Q-Network algorithm for RL
"""
class DQN(DecisionMaker):

    def __init__(self, input_dims, num_actions, is_conv=False, learning_rate=3e-4, gamma=0.99,
                 buffer_size=10000, mapping_fn=None):
        super().__init__()
        self.input_dims = input_dims[::-1]
        self.num_actions = num_actions
        self.gamma = gamma
        self.replay_buffer = BasicBuffer(max_size=buffer_size)
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        if is_conv:
            self.model = ConvDQNetwork(self.input_dims, self.num_actions).to(self.device)
        else:
            self.model = DQNetwork(self.input_dims, self.num_actions, mapping_fn).to(self.device)

        self.optimizer = torch.optim.Adam(self.model.parameters(), lr=learning_rate)
        self.MSE_loss = nn.MSELoss()

    def get_action(self, observation):
        """Get an action according the network"""
        observation = autograd.Variable(torch.from_numpy(observation.copy()).float().unsqueeze(0))
        qvals = self.model.forward(observation)
        action = np.argmax(qvals.cpu().detach().numpy())

        return action

    def get_train_action(self, observation, eps=0.1):
        """Get action for when training"""
        observation = autograd.Variable(torch.from_numpy(observation.copy()).float().unsqueeze(0))
        qvals = self.model.forward(observation)
        action = np.argmax(qvals.cpu().detach().numpy())

        if np.random.randn() < eps:
            return random.randint(0, self.num_actions - 1)

        return action

    def compute_loss(self, batch):
        states, actions, rewards, next_states, dones = batch
        states = torch.FloatTensor(states)
        actions = torch.LongTensor(actions)
        rewards = torch.FloatTensor(rewards)
        next_states = torch.FloatTensor(next_states)
        dones = torch.FloatTensor(dones)

        curr_Q = self.model.forward(states).gather(1, actions.unsqueeze(1))
        curr_Q = curr_Q.squeeze(1)
        next_Q = self.model.forward(next_states)
        max_next_Q = torch.max(next_Q, 1)[0]
        expected_Q = rewards.squeeze(1) + (1 - dones) * self.gamma * max_next_Q

        loss = self.MSE_loss(curr_Q, expected_Q.detach())
        return loss

    def update_step(self, obs, action,new_obs, reward, done):
        """Add the step to the replay buffer"""
        self.replay_buffer.push(obs, action, new_obs, reward, done)

    def update_episode(self, batch_size):
        """Updating the Network"""
        if len(self.replay_buffer) > batch_size:
            batch = self.replay_buffer.sample(batch_size)
            loss = self.compute_loss(batch)
            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()


"""Class for the fully-connected Q-Network"""
class DQNetwork(nn.Module):

    def __init__(self, input_dim, output_dim, mapping_fn=None):
        super(DQNetwork, self).__init__()
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.mapping_fn = mapping_fn
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        self.fc = nn.Sequential(
            nn.Linear(self.input_dim[0], 128),
            nn.ReLU(),
            nn.Linear(128, 256),
            nn.ReLU(),
            nn.Linear(256, self.output_dim)
        )

    def forward(self, observation):
        if self.mapping_fn:
            observation = self.mapping_fn(observation)

        qvals = self.fc(observation)
        return qvals


"""Class for the convolution Q-Network"""
class ConvDQNetwork(nn.Module):

    def __init__(self, input_dim, output_dim):
        super(ConvDQNetwork, self).__init__()
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        self.conv = nn.Sequential(
            nn.Conv2d(self.input_dim[0], 32, kernel_size=8, stride=1),
            nn.ReLU(),
            nn.Conv2d(32, 64, kernel_size=4, stride=2),
            nn.ReLU(),
            nn.Conv2d(64, 64, kernel_size=3, stride=1),
            nn.ReLU()
        )

        self.fc_input_dim = self.feature_size()

        self.fc = nn.Sequential(
            nn.Linear(self.fc_input_dim, 128),
            nn.ReLU(),
            nn.Linear(128, 256),
            nn.ReLU(),
            nn.Linear(256, self.output_dim)
        )

    def forward(self, observation):
        observation = observation.permute(0, 3, 1, 2).to(self.device)
        features = self.conv(observation)
        features = features.view(features.size(0), -1)
        qvals = self.fc(features)
        return qvals

    def feature_size(self):
        return self.conv(autograd.Variable(torch.zeros(1, *self.input_dim))).view(1, -1).size(1)

"""Class for the Replay Buffer"""
class BasicBuffer:

    def __init__(self, max_size):
        self.max_size = max_size
        self.buffer = deque(maxlen=max_size)

    def push(self, state, action, reward, next_state, done):
        experience = (state, action, np.array([reward]), next_state, done)
        self.buffer.append(experience)

    def sample(self, batch_size):
        state_batch = []
        action_batch = []
        reward_batch = []
        next_state_batch = []
        done_batch = []

        batch = random.sample(self.buffer, batch_size)

        for experience in batch:
            state, action, reward, next_state, done = experience
            state_batch.append(state)
            action_batch.append(action)
            reward_batch.append(reward)
            next_state_batch.append(next_state)
            done_batch.append(done)

        return state_batch, action_batch, reward_batch, next_state_batch, done_batch

    def __len__(self):
        return len(self.buffer)


import time

import numpy as np
from numpy.core.arrayprint import _get_format_function
from environments.corners.corners_env import CornersEnv, Action

class GoalAgent:
    """Move toward a goal, 50% for move in x
    """
    def __init__(self, goal):
        self.goal_y, self.goal_x = goal

    def get_action(self, observation):
        y, x = observation[0, :]
        actions = []

        if x > self.goal_x:
            actions.append(Action.left)
        elif x < self.goal_x:
            actions.append(Action.right)

        if y > self.goal_y:
            actions.append(Action.up)
        elif y < self.goal_y:
            actions.append(Action.down)
        
        if not actions:
            # If we reached the goal
            actions.append(Action.noop)

        return np.random.choice(actions)

class EconomicGoalAgent(GoalAgent):
    """Like Goal agent, also also bids and the goal
    can be set later
    """
    def __init__(self, position=[0, 0]):
        super().__init__([0,0])
        self.pos = position # We need to know to position for the bidding:

    def reset(self, observation):
        self.pos = observation[0, :]

    def get_bid(self, suggested_goal):
        # Manhatten Distance
        return abs(self.pos[0] - suggested_goal[0]) + abs(self.pos[1] - suggested_goal[1])

    def set_task(self, task):
        self.goal_y, self.goal_x = task

    

def main():
    # Test the agents:
    env = CornersEnv()
    obs = env.reset()
    goal_agents = {
        'A': GoalAgent([0, 0]),
        'B': GoalAgent([0, 4]),
        'C': GoalAgent([4, 0]),
        'D': GoalAgent([4, 4]),
        'E': GoalAgent([2, 2])
    }
    done = {'temp_agent': False}

    while(not all(value == True for value in done.values())):
        actions = {agent: goal_agents[agent].get_action(obs[agent]) for agent in env.agents}
        obs, _, done, _ = env.step(actions)
        env.render()
        time.sleep(0.2)


if __name__ == "__main__":
    main()


class EnvWrappper:

    def __init__(self, env, needs_conv=False):
        self.env = env
        self.num_obs = self.env.observation_spaces[env.possible_agents[0]].shape
        self.num_actions = self.env.action_spaces[env.possible_agents[0]].n
        self.needs_conv = needs_conv
        self.env_agents = self.env.possible_agents

    def get_env(self):
        return self.env

    def get_num_obs(self):
        return self.num_obs

    def get_num_actions(self):
        return self.num_actions

    def get_needs_conv(self):
        return self.needs_conv

    def get_env_agents(self):
        return self.env_agents

    def step(self, joint_action):
        return self.env.step(joint_action)



import numpy as np
from environments.cleanup.social_dilemmas.envs.cleanup import CleanupEnv

env = CleanupEnv(num_agents=5, render=True)

agents = list(env.agents.values())
print(agents)

for i in range(100):
    action_dim = agents[0].action_space.n
    rand_action = np.random.randint(action_dim, size=5)
    obs, rew, dones, info, = env.step({'agent-0': rand_action[0],
                                       'agent-1': rand_action[1],
                                       'agent-2': rand_action[2],
                                       'agent-3': rand_action[3],
                                       'agent-4': rand_action[4]})
    from pprint import pprint; pprint(obs['agent-4'].shape)
    exit()

    # print(rew, dones)
    env.render()


class Agent:

    def __init__(self, decision_maker):
        self.decision_maker = decision_maker

    def get_decision_maker(self):
        return self.decision_maker


"""
An abstract class for choosing an action, part of an agent.
(An agent can have one or several of these)
"""
class DecisionMaker:

    def __init__(self):
        pass

    def get_action(self, observation):
        pass

    """
    Functions for training:
    """
    def get_train_action(self, observation):
        pass

    def update_step(self, obs, action,new_obs, reward, done):
        pass

    def update_episode(self, batch_size=0):
        pass
